//////////////////////////////////////////
//
//  compare_maf_v5.c implemented by Minmei Hou
//  		     inital implementation in May 2004
// 		     modified in June,July 2004
//                   prolog added in Aug. 2004
// 
//  Please send bug to mhou@cse.psu.edu
//
//////////////////////////////////////////
//  
//  The program comparing two alignments files 
//  which follow either maf format or fasta format.
//
//  The program accepts two arguments: two file
//  names to be compared. If there is a "true" 
//  alignment among the the two files, it should 
//  be the second file. 
// 
//  The program outputs counts and percentages
//  useful for evaluation purpose: agreement, 
//  sensitivity, specificity, agreement_5. Among 
//  them, sensitity and specificity are meaningful
//  when the second file is "true" alignment. 
//
//  From output -- 
//  agreement1: fraction of positions of the first alignment 
//              file that agree with the second file.               
//  agreement2: fraction of positions of the second alignment
//              file that agree with the first file.
//  specificity: fraction of aligned positions of the first alignment
//               file that are the same as in the second file.
//  sensitivity: fraction of aligned positions of the second alignment
//               file that are the same as in the first file.
//  agreement_5: similar to agreement1, but the "agreement" includes
//               mismatches within 5 bases.
//
//
//  For this version, counts for positive and 
//  negative orientation are not seperated. The 
//  values show the combination of both. 
//
//  The alignment difference is output to 
//  "compare.diff". 
//  
//  Limits: this program requires sequence files.
//  Each sequence length is below 1,000,000.
//
///////////////////////////////////////////
//
//  source files required for compilation:
//  maf.c/h, util.c/h, seq.c/h, tba_util.c/h, 
//  nib.c/h, charvec.c/h, compare_maf_v5.c    
//
//  gcc -g -Wall compare_maf_v5.c maf.c util.c
//  seq.c tba_util.c nib.c charvec.c -o compare_maf_v5
///////////////////////////////////////////

// collect agreement, sensitivity, specificity, fpr, but regardless of positive or negative
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "maf.h"
#include "tba_util.h"
#include "seq.h"

#define MAX_SEQLEN 1000000
#define MAX_SPIECES 1000
#define MAX_NAMESIZE 100

static int distanceVector[MAX_SEQLEN];
static int largest;      // record largest distance difference, for output purpose
static char* refName;
static char* seqName;
static char* curRefName;
static char* curSeqName;
static int refSize, seqSize, Tpos, T5pos, Mpos, Wori, Fpos, Fneg, Tneg, true_aligned, pred_aligned;

struct retPair {
  float* rankArray;
  char*  eleArray1;
  char*  eleArray2;
  char*  strandArray;
};

void strip(char *s)
{
  int j=0;
  int l=strlen(s);
  int i;
  for (i=0;i<l;i++)
    {
      //      if (!isspace(s[i])) {s[j]=tolower(s[i]);j++;}
      if ( !isspace(s[i])) {s[j]=s[i];j++;}
    }
  s[j]='\0';
}

void printPartArray(struct retPair *a, struct retPair *b, int beg, int end, FILE* fp, char* name1, char* name2) {
  struct retPair * retpairArray[2], *ptr;
  int i,j,size, start, tail;
  char *s1, *s2;

  fprintf(fp, "a score=\n");
  retpairArray[0] = a;
  retpairArray[1] = b;

  for (j=0; j<2; j++) { // from first file and second file
    ptr = retpairArray[j];
    s1 = ptr->eleArray1;
    s2 = ptr->eleArray2;
    size = 0;
    start = tail = -1; // size, start, tail are records for sequence 2
    fprintf(fp, "s %s\t%d\t%d\t%c %d\t", name1, beg, end-beg+1, '+', refSize);  
    for (i=beg; i<=end; i++) {
      fprintf(fp, "%c", s1[i]);
      if ( ptr->rankArray[i] == (int)(ptr->rankArray)[i]) {
        size++;  // size is the size of fragment from sequence 2
	if ( start == -1)
	  start = (int)(ptr->rankArray)[i];
	tail = (int)(ptr->rankArray)[i];
      }
    }
    fprintf(fp, "\n");
    // fragment length of sequence2 is counted by adding non 0.5 or -1 elements from beg to end(including beg, end, as above size
    if ( start != -1)
      fprintf(fp, "s %s\t%d\t%d\t%c %d\t", name2, start, size, retpairArray[j]->strandArray[beg], seqSize);   
    else {
      start = (int)(ptr->rankArray)[beg];
      fprintf(fp, "s %s\t%d\t0\t%c %d\t", name2, start, retpairArray[j]->strandArray[beg], seqSize);
    }
    for (i=beg; i<=end; i++)
      fprintf(fp, "%c", s2[i]);
    fprintf(fp, "\n");
  }

  fprintf(fp, "\n");
}

//--------< Collect rank values for alignment file in fasta format >----------------------
// char **fasta is an array of ptrs to sequences
// char **namelist is an array of ptrs to names
// char *name1, *name2 are two species to be extracted, name1 is ref.
//
struct retPair* fasta2Array(char **fasta, char** namelist, char* name1, char* name2) {
  int i, j, len, count1, count2;
  char *s, *t;
  float* rankarray;
  char *elearray1, *elearray2, *strandArray;
  struct retPair *retpair;
  
  for (i=0; i<MAX_SEQLEN && strcmp(namelist[i], name1)!=0; i++)
    ;
  for (j=0; j<MAX_SEQLEN && strcmp(namelist[j], name2)!=0; j++)
    ;
  if ( i==MAX_SEQLEN || j==MAX_SEQLEN)  // name1 or name2 not found in namelist
    return NULL;
  s = fasta[i]; // sequence 1
  t = fasta[j]; // sequence 2

  for (len=i=0; s[i]; i++) {
    if ( s[i] != '-')
      len++;
  }
  rankarray = (float*)malloc(len*sizeof(float));
  elearray1 = (char*)malloc(len*sizeof(char));
  elearray2 = (char*)malloc(len*sizeof(char));
  strandArray = (char*)malloc(len*sizeof(char));

  for (i=count1=count2=0; s[i]; i++) {
    if (s[i] != '-') {
      elearray1[count1] = s[i];
      elearray2[count1] = t[i];
      strandArray[count1] = '+';
      if (t[i] != '-') 
	rankarray[count1] = count2;
      else
	rankarray[count1] = count2 - 0.5;
      count1++;
    }
    if (t[i] != '-')
      count2++;
  }
  retpair = (struct retPair*)malloc(sizeof(struct retPair));
  retpair->rankArray = rankarray;
  retpair->eleArray1 = elearray1;
  retpair->eleArray2 = elearray2;
  retpair->strandArray = strandArray;
  return retpair;
}

//----------< Collect rank values for alignment file in maf format >-----
// char* filename is the maf filename, assume it's maf formate
// char* name1, *name2 are two species to be extracted
// char orient is +/-
// int srcSize1 is the length of sequence1
//
struct retPair* maf2Array(char* filename, char* name1, char* name2, int srcSize1) {
  int i, pos, seq_len, count_ref, count_seq;
  char *s, *t, *elearray1, *elearray2, *strandArray;
  float* rankarray;
  struct mafFile *mf;
  struct mafAli *ali;
  struct mafComp* comp1, *comp2;
  struct retPair* retpair;
  SEQ* seq1;
  int seq1_len;
  char *strptr;

  rankarray = (float*)malloc(srcSize1*sizeof(float));
  elearray1 = (char*)malloc(srcSize1*sizeof(char));
  elearray2 = (char*)malloc(srcSize1*sizeof(char));
  strandArray = (char*)malloc(srcSize1*sizeof(char));
  for (i=0; i<srcSize1; i++) {
    rankarray[i] = -0.5;
    elearray2[i] = '-'; // by default
    strandArray[i] = '?';   // by default
  }
  seq1 = seq_get(name1);
  seq1_len = SEQ_LEN(seq1);
  for (i=0, strptr = SEQ_CHARS(seq1); i<seq1_len; i++)
    elearray1[i] = strptr[i];
  seq_close(seq1);

  mf = mafOpen(filename, 0);
  while ((ali = mafNext(mf)) != NULL) {
    comp1 = ali->components;
    while (comp1 != NULL && strcmp(comp1->src, name1)!=0)
      comp1 = comp1->next;
    if (comp1 == NULL) {
      mafAliFree(&ali);
      continue;
    }
    if ( comp1->strand == '-') {  // make sure reference sequence is positive orient
      rc(ali);
      comp1 = ali->components;
      while (comp1 != NULL && strcmp(comp1->src,name1)!=0)
	comp1 = comp1->next;
    }
    comp2 = ali->components;
    while (comp2 != NULL && strcmp(comp2->src, name2)!=0) 
      comp2 = comp2->next;
    if ( comp2 == NULL) {
      mafAliFree(&ali);
      continue;
    }
      s = comp1->text;
      t = comp2->text;     
      seq_len = strlen(s);
      for (pos=count_ref=count_seq = 0; pos<seq_len; pos++) {
      	if ( s[pos] != '-') {
	  elearray2[comp1->start + count_ref] = t[pos]; 
	  strandArray[comp1->start + count_ref] = comp2->strand;
	  if ( t[pos] != '-') {
	    if ( comp2->strand == '+')
	      rankarray[comp1->start + count_ref] = comp2->start + count_seq;
	    else
	      rankarray[comp1->start + count_ref] = comp2->srcSize - (comp2->start + count_seq) - 1;
	  }
	  count_ref++;
        } 
        if ( t[pos] != '-')   		
          count_seq++;
      }
    mafAliFree(&ali);
  }
  mafFileFree(&mf);
  retpair = (struct retPair*)malloc(sizeof(struct retPair));
  retpair->rankArray = rankarray;
  retpair->eleArray1 = elearray1;
  retpair->eleArray2 = elearray2;
  retpair->strandArray = strandArray;
  return retpair;
}

//--------< for purpose to output difference >--------------------
// 
void processPosition(struct retPair *a, struct retPair *b, int j, int* start, int* last, FILE* fp) {
  if (j == *last + 1)
    *last = j;
  else{
    if (*last >= 0)
      printPartArray(a, b, *start, *last, fp, curRefName, curSeqName);
    *start = *last = j;
  }
}

void freeRetPair(struct retPair** retpair) {
  free((*retpair)->eleArray1);
  free((*retpair)->eleArray2);
  free((*retpair)->rankArray);
  free(*retpair);
  *retpair = NULL;
}

// Tpos, Mpos, Fpos, Fneg, Tneg, true_aligned, pred_aligned, 
//-------< Evaluate between two retPair structures for a pair of species> ---------------------
// 1a, 2a use same sequence as reference, the original reference
void processSeqPairRef(struct retPair *retpair1a, struct retPair *retpair2a, FILE* fp) {
  float *rank1a, *rank2a;
  int i, tmp, start, last;
  char *non_ref1, *non_ref2;

  if ( retpair1a == NULL || retpair2a == NULL)
    return;  
  rank1a = retpair1a->rankArray;
  rank2a = retpair2a->rankArray;
  start = last = -100;

  for (i=0; i<refSize; i++) {
    if ( rank1a[i] == (int)rank1a[i]) { // maf1 x/x
      pred_aligned++;
      if ( rank2a[i] == (int)rank2a[i]) { // maf2 x/x
	true_aligned++;
	if ( retpair1a->strandArray[i]==retpair2a->strandArray[i] ) {
	  if ( (int)rank1a[i] == (int)rank2a[i] ) { // maf2 x/x
	    Tpos++;  // maf1 x/x, maf2 x/x, same
	    T5pos++;
	  }
	  else {
	    tmp = rank1a[i] > rank2a[i] ? rank1a[i]-rank2a[i] : rank2a[i] - rank1a[i];
	    distanceVector[(int)tmp]++;
	    largest = largest > tmp ? largest : tmp;
	    Mpos++;  // maf1 x/x, maf2 x/x, not same
	    if ( tmp <= 5 )
	      T5pos++;
	    processPosition(retpair1a, retpair2a, i, &start, &last, fp);
	  }
	}
	else {
	  Wori++;
	  processPosition(retpair1a, retpair2a, i, &start, &last, fp);
	}
      }
      else {
	Fpos++; // maf1 x/x, maf2 x/-
	processPosition(retpair1a, retpair2a, i, &start, &last, fp);
      }
    }
    else { // maf1 x/-
      if ( rank2a[i] == (int)rank2a[i]) { // maf2 x/x
	true_aligned++;
	Fneg++; // maf1 x/-, maf2 x/x
	processPosition(retpair1a, retpair2a, i, &start, &last, fp);
      }
      else // maf2 x/-
	Tneg++;
    }
  }
  if ( last >= 0)
    printPartArray(retpair1a, retpair2a, start, last, fp, refName, seqName);

  // need to compute Tneg in respect to non-ref sequence
  non_ref1 = (char*)malloc(seqSize*sizeof(char));
  non_ref2 = (char*)malloc(seqSize*sizeof(char));
  for (i=0; i<seqSize; i++)
    non_ref1[i] = non_ref2[i] = '-';
  for (i=0; i<refSize; i++) {
    if (rank1a[i] == (int)rank1a[i]) // maf1 x/x
      non_ref1[(int)rank1a[i]] = '*';
    if (rank2a[i] == (int)rank2a[i]) // maf2 x/x
      non_ref2[(int)rank2a[i]] = '*';
  }
  for (i=0; i<seqSize; i++) {
    if ( non_ref1[i] == '-' && non_ref2[i] == '-')
      Tneg++; 
  }
}

char** collectMafNamesAndSizes(char* filename, char** namelist, int sizes[MAX_SPIECES]) {
  int i, specie_number;
  struct mafFile* mf;
  struct mafAli* ali;
  struct mafComp* mc;
  
  specie_number = 0;
  mf = mafOpen(filename, 0);
  while ((ali = mafNext(mf)) != NULL) {
    mc = ali->components;
    while ( mc != NULL) {
      for(i=0;namelist[i]!=NULL;i++) {
        if (strcmp(mc->src, namelist[i]) == 0)
          break;
      }
      if ( namelist[i] == NULL) {
        namelist[i] = (char*)malloc((strlen(mc->src)+1)*sizeof(char));
        strcpy(namelist[i], mc->src);
        sizes[i] = mc->srcSize;
      }
      mc = mc->next;
    }
  }
  mafFileFree(&mf);
  return namelist;
}

char** collectFastaNamesAndSeqsAndSizes(char* filename, char** namelist, char** seqs, int sizes[MAX_SPIECES]) {
  int nb_seq = 0;
  char *s = (char*)malloc(MAX_SEQLEN*sizeof(char));
  char nm[MAX_NAMESIZE];
  FILE *f = fopen(filename, "r");
  int i, len;

  if ( f==NULL) {
    printf("cna't open file\n");
    return NULL;
  }
  while (fscanf(f, ">%s\n%[^>]", nm, s) != EOF) {
    seqs[nb_seq] = (char*)malloc(strlen(s)+1);
    strip(s);
    strcpy(seqs[nb_seq], s);
    nm[MAX_NAMESIZE-1] = '0';
    namelist[nb_seq] = (char*)malloc(sizeof(char)*(strlen(nm)+1));
    strcpy(namelist[nb_seq], nm);
    for(i=len=0; s[i]!='\0'; i++)
      if ( s[i] != '-')
	len++;
    sizes[nb_seq] = len;
    nb_seq++;
  }
  fclose(f);
  free(s);
  return namelist;
}

// refSize, seqSize, Tpos, Mpos, Fpos, Fneg, Tneg, true_aligned, pred_aligned;
void printStat(int *disVec, int refsize, int seqsize, int tpos, int t5pos, int mpos, int wori, int fpos, int fneg, int tneg, int true_ali, int pred_ali) {
  int m;
  printf("refSize(%d) seqSize(%d) pred_aligned(%d) true_aligned(%d) length_align1(%d) length_align2(%d) Tpos(%d) T5pos(%d) Mpos(%d) Wori(%d) Fpos(%d) Fneg(%d) Tneg(%d)\n", refsize, seqsize, pred_ali, true_ali, refsize+seqsize-pred_ali, refsize+seqsize-true_ali, tpos, t5pos, mpos, wori, fpos, fneg, tneg);
  printf("agreement1(%5.3f)\tagreement2(%5.3f)\tspecificity(%5.3f)\tsensitivity(%5.3f)\tfpr(%5.3f)\tagreement_5(%5.3f)\n", 100*(tpos+tneg)/(float)(refsize+seqsize-pred_ali), 100*(tpos+tneg)/(float)(refSize+seqSize-true_ali), 100*tpos/(float)(pred_ali), 100*tpos/(float)(true_ali), 100*fpos/(float)(refsize - true_ali), 100*(t5pos+tneg)/(float)(refsize+seqsize-pred_ali));

  for (m=0;m<=largest;m++) 
    if ( disVec[m] > 0)
      printf("%d: %5.1f(%d)\t", m, 100*disVec[m]/(float)mpos, disVec[m]);
  printf("\n");	
}	

int main(int argc, char *argv[]) {
  int maf1, maf2, sizelist1[MAX_SPIECES], sizelist2[MAX_SPIECES], sizelist[MAX_SPIECES], tDistanceVector[MAX_SEQLEN];
  int tRefSize, tSeqSize, tTpos, tT5pos, tMpos, tWori, tFpos, tFneg, tTneg, tTrue_aligned, tPred_aligned;
  int i,j,m, k;
  char **namelist1, **namelist2, **namelist, **seqs1, **seqs2;
  FILE *fp;
  struct retPair *retpair1a, *retpair2a;

  //--------------- check input format ------------------------//
  if ( argc != 3) {
    printf("Need two files as argument...\n");
    return 1;
  }

  if (strstr(argv[1], ".maf") != NULL)
    maf1 = 1;
  else if (strstr(argv[1], ".fa") != NULL)
    maf1 = 0;
  else {
    printf("First argument wrong format...\n");
    return 1;
  }

  if (strstr(argv[2], ".maf") != NULL)
    maf2= 1;
  else if (strstr(argv[2], ".fa") != NULL)
    maf2 = 0;
  else {
    printf("Second argument wrong format...\n");
    return 1;
  }
  //--------------- collect names,sizes,seqs ----------------------// 
  refName = (char*)malloc(MAX_NAMESIZE*sizeof(char));
  seqName = (char*)malloc(MAX_NAMESIZE*sizeof(char));
  curRefName = (char*)malloc(MAX_NAMESIZE*sizeof(char));
  curSeqName = (char*)malloc(MAX_NAMESIZE*sizeof(char));
  namelist1 = (char**)malloc(MAX_SPIECES*sizeof(char*));
  namelist2 = (char**)malloc(MAX_SPIECES*sizeof(char*));
  namelist  = (char**)malloc(MAX_SPIECES*sizeof(char*));
  seqs1 = (char**)malloc(MAX_SPIECES*sizeof(char*));
  seqs2 = (char**)malloc(MAX_SPIECES*sizeof(char*));
  
  for(i=0;i<MAX_SPIECES;i++) {
    namelist1[i] = namelist2[i] = namelist[i] = seqs1[i] = seqs2[i] = NULL;
    sizelist1[i] = sizelist2[i] = sizelist[i] = 0;
  }
  if (maf1==1)
    collectMafNamesAndSizes(argv[1], namelist1, sizelist1);
  else
    collectFastaNamesAndSeqsAndSizes(argv[1], namelist1, seqs1, sizelist1);
  if (maf2==1)
    collectMafNamesAndSizes(argv[2], namelist2, sizelist2);
  else 
    collectFastaNamesAndSeqsAndSizes(argv[2], namelist2, seqs2, sizelist2);

  for (i=k=0; namelist1[i]!=NULL && i<MAX_SPIECES;i++) { // namelist holds species in common, sizelist holds smaller value of two
    for(j=0; namelist2[j]!=NULL && j<MAX_SPIECES;j++) {
      if (strcmp(namelist1[i],namelist2[j])==0)
        break;
    }
    if ( namelist2[j] != NULL) {
      namelist[k] = (char*)malloc((MAX_NAMESIZE+1)*sizeof(char));
      strcpy(namelist[k], namelist1[i]);
      sizelist[k] = sizelist1[i] < sizelist2[j] ? sizelist1[i] : sizelist2[j];
      k++;
    }
  }
  //---------------------- start to process and print ---------//
  fp = fopen("compare.diff", "w");
  fprintf(fp, "##maf version=1 scoring=\n");
  tRefSize = tSeqSize = tTpos = tT5pos = tMpos = tWori = tFpos = tFneg = tTneg = tTrue_aligned = tPred_aligned = 0;

  for (m=0; m<MAX_SEQLEN; m++)
    tDistanceVector[m] = 0;
  largest=-1;

  for(i=0;namelist[i]!= NULL;i++) {
    for(j=i+1;namelist[j]!=NULL;j++) {
      if ( i!=j) { // for each pair
        strcpy(refName, namelist[i]);
	strcpy(seqName, namelist[j]);
	refSize = sizelist[i];
	seqSize = sizelist[j];

  	for (m=0; m<MAX_SEQLEN; m++)
    	  distanceVector[m] = 0;
	Tpos = T5pos = Mpos = Wori = Fpos = Fneg = Tneg = true_aligned = pred_aligned = 0;

        printf("%s & %s\t", namelist[i], namelist[j]);
        
	if (maf1==1) 
	  retpair1a = maf2Array(argv[1], namelist[i], namelist[j], sizelist[i]);  // top sequence as reference
	else
	  retpair1a = fasta2Array(seqs1, namelist1, namelist[i], namelist[j]);
	if (maf2==1)
	  retpair2a = maf2Array(argv[2], namelist[i], namelist[j], sizelist[i]);
	else
	  retpair2a = fasta2Array(seqs2, namelist2, namelist[i], namelist[j]);
	
	strcpy(curRefName, refName);
	strcpy(curSeqName, seqName);
	processSeqPairRef(retpair1a, retpair2a, fp);
	freeRetPair(&retpair1a);
	freeRetPair(&retpair2a);
      }
      printStat(distanceVector, refSize, seqSize, Tpos, T5pos, Mpos, Wori, Fpos, Fneg, Tneg, true_aligned, pred_aligned);
      for (m=0;m<=largest;m++) 
	tDistanceVector[m] += distanceVector[m];
      tRefSize += refSize;
      tSeqSize += seqSize;
      tTpos += Tpos;
      tT5pos += T5pos;
      tMpos += Mpos;
      tFpos += Fpos;
      tFneg += Fneg;
      tTneg += Tneg;
      tTrue_aligned += true_aligned;
      tPred_aligned += pred_aligned;
    }
  }

  printf("------------------------< Overall >-------------------\n");
  printStat(tDistanceVector, tRefSize, tSeqSize, tTpos, tT5pos, tMpos, tWori, tFpos, tFneg, tTneg, tTrue_aligned, tPred_aligned);
  fclose(fp);

  for (i=0; i<MAX_SPIECES; i++) {
    if ( seqs1[i] != NULL)
      free(seqs1[i]);
    if ( seqs2[i] != NULL)
      free(seqs2[i]);
  }
  free(seqs1);
  free(seqs2);

  for (i=0; namelist1[i] != NULL && i<MAX_SPIECES; i++) {
    free(namelist1[i]);
    free(namelist2[i]);
    free(namelist[i]);
  }
  free(namelist1);
  free(namelist2);
  free(namelist);
  return 0;
}
